# Generated by Django 4.2.16 on 2024-12-12 08:01

from django.db import migrations, models

import resolwe.flow.models.base


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0033_data_idx_data_status_priority"),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="data",
            name="idx_data_status_priority",
        ),
        migrations.AddIndex(
            model_name="data",
            index=models.Index(
                models.Case(
                    models.When(
                        status=resolwe.flow.models.base.DataStatus["ERROR"],
                        then=models.Value(0),
                    ),
                    models.When(
                        status=resolwe.flow.models.base.DataStatus["PROCESSING"],
                        then=models.Value(1),
                    ),
                    models.When(
                        status=resolwe.flow.models.base.DataStatus["UPLOADING"],
                        then=models.Value(2),
                    ),
                    models.When(
                        status=resolwe.flow.models.base.DataStatus["PREPARING"],
                        then=models.Value(3),
                    ),
                    models.When(
                        status=resolwe.flow.models.base.DataStatus["WAITING"],
                        then=models.Value(4),
                    ),
                    models.When(
                        status=resolwe.flow.models.base.DataStatus["RESOLVING"],
                        then=models.Value(5),
                    ),
                    models.When(
                        status=resolwe.flow.models.base.DataStatus["DONE"],
                        then=models.Value(6),
                    ),
                ),
                name="idx_data_status_priority",
            ),
        ),
    ]



""".. Ignore pydocstyle D400.

.. autoclass:: resolwe.flow.executors.docker.run.FlowExecutor
    :members:

"""

# pylint: disable=logging-format-interpolation
import asyncio
import copy
import functools
import logging
import os
import tempfile
import time
from contextlib import suppress
from pathlib import Path
from typing import Any, Dict, Iterable, Tuple, Type

import docker

from .. import constants
from ..connectors import connectors
from ..connectors.baseconnector import BaseStorageConnector
from ..global_settings import LOCATION_SUBPATH, PROCESS_META, SETTINGS
from ..local.run import FlowExecutor as LocalFlowExecutor
from ..protocol import ExecutorFiles

# Limits of containers' access to memory. We set the limit to ensure
# processes are stable and do not get killed by OOM signal.
DOCKER_MEMORY_HARD_LIMIT_BUFFER = 100
DOCKER_MEMORY_SWAP_RATIO = 2
DOCKER_MEMORY_SWAPPINESS = 1

# Secrets necessary to connect to the listener service.
LISTENER_PUBLIC_KEY = os.getenv("LISTENER_PUBLIC_KEY")
PUBLIC_KEY = os.getenv("CURVE_PUBLIC_KEY")
PRIVATE_KEY = os.getenv("CURVE_PRIVATE_KEY")

logger = logging.getLogger(__name__)


def retry(
    max_retries: int = 3,
    retry_exceptions: Tuple[Type[Exception], ...] = (
        docker.errors.ImageNotFound,
        docker.errors.APIError,
    ),
    min_sleep: int = 1,
    max_sleep: int = 10,
):
    """Try to call decorated method max_retries times before giving up.

    The calls are retried when function raises exception in retry_exceptions.

    :param max_retries: maximal number of calls before giving up.
    :param retry_exceptions: retry call if one of these exceptions is raised.
    :param min_sleep: minimal sleep between calls (in seconds).
    :param max_sleep: maximal sleep between calls (in seconds).
    :returns: return value of the called method.
    :raises: the last exceptions raised by the method call if none of the
      retries were successfull.
    """

    def decorator_retry(func):
        @functools.wraps(func)
        def wrapper_retry(*args, **kwargs):
            last_error: Exception = Exception("Retry failed")
            sleep: int = 0
            for retry in range(max_retries):
                try:
                    time.sleep(sleep)
                    return func(*args, **kwargs)
                except retry_exceptions as err:
                    sleep = min(max_sleep, min_sleep * (2**retry))
                    last_error = err
            raise last_error

        return wrapper_retry

    return decorator_retry


def get_mountable_connectors() -> Iterable[Tuple[str, BaseStorageConnector]]:
    """Iterate through all the storages and find mountable connectors.

    :returns: list of tuples (storage_name, connector).
    """
    return [
        (storage_name, connector)
        for storage_name in SETTINGS["FLOW_STORAGE"]
        for connector in connectors.for_storage(storage_name)
        if connector.mountable
    ]


def get_upload_dir() -> str:
    """Get the upload path.

    : returns: the path of the first mountable connector for storage
        'upload'.

    :raises RuntimeError: if no applicable connector is found.
    """
    for connector in connectors.for_storage("upload"):
        if connector.mountable:
            return f"/upload_{connector.name}"
    raise RuntimeError("No mountable upload connector is defined.")


def unique_volume_name(base_name: str, data_id: int) -> str:
    """Get unique persistent volume claim name."""
    return f"{base_name}-{data_id}"


class FlowExecutor(LocalFlowExecutor):
    """Docker executor."""

    name = "docker"

    def __init__(self, *args, **kwargs):
        """Initialize attributes."""
        super().__init__(*args, **kwargs)
        container_name_prefix = SETTINGS.get("FLOW_EXECUTOR", {}).get(
            "CONTAINER_NAME_PREFIX", "resolwe"
        )
        self.container_name = self._generate_container_name(container_name_prefix)
        self.tools_volumes = []
        self.command = SETTINGS.get("FLOW_DOCKER_COMMAND", "docker")
        self.tmpdir = tempfile.TemporaryDirectory()
        self.processing_working_dir = constants.PROCESSING_VOLUME

    def _new_volume(
        self, config: Dict[str, Any], mount_path: Path, read_only: bool = True
    ) -> Tuple[str, Dict[str, str]]:
        """Generate a new volume entry.

        :param config: must include 'path' and may include 'selinux_label' and
            'docker_volume'.
        :param mount_moint: mount point for the volume.
        """
        options = set()
        if "selinux_label" in config:
            options.add(config["selinux_label"])
        options.add("ro" if read_only else "rw")
        # First try to mount is as volume, if fails try the path.
        volume_path = config.get("docker_volume", config["path"])
        return (
            os.fspath(volume_path),
            {"bind": os.fspath(mount_path), "mode": ",".join(options)},
        )

    def _get_volumes(self, subpaths=False) -> Dict[str, Tuple[Dict, Path]]:
        """Get writeable volumes from settings.

        :attr subpaths: when True the location subpath in added to the volume
            path.

        :returns: mapping between volume name and tuple (config, mount_point).
        """
        results = dict()
        volume_mountpoint = {
            constants.PROCESSING_VOLUME_NAME: constants.PROCESSING_VOLUME,
            constants.INPUTS_VOLUME_NAME: constants.INPUTS_VOLUME,
            constants.SECRETS_VOLUME_NAME: constants.SECRETS_VOLUME,
            constants.SOCKETS_VOLUME_NAME: constants.SOCKETS_VOLUME,
        }

        for volume_name, volume in SETTINGS["FLOW_VOLUMES"].items():
            if "read_only" not in volume["config"]:
                if volume["type"] == "host_path":
                    config = copy.deepcopy(volume["config"])
                    if subpaths:
                        config["path"] = Path(config["path"]) / LOCATION_SUBPATH
                    results[volume_name] = (config, volume_mountpoint[volume_name])
                # The persistent volume in Docker is actually named volume.
                elif volume["type"] == "persistent_volume":
                    config = copy.deepcopy(volume["config"])
                    config["path"] = unique_volume_name(config["path"], self.data_id)
                    results[volume_name] = (config, volume_mountpoint[volume_name])
                    if subpaths:
                        self.processing_working_dir /= LOCATION_SUBPATH
                elif volume["type"] == "temporary_directory":
                    config = copy.deepcopy(volume["config"])
                    volume_path = Path(self.tmpdir.name) / volume_name
                    mode = config.get("mode", 0o700)
                    volume_path.mkdir(exist_ok=True, mode=mode)
                    config["path"] = volume_path
                    results[volume_name] = (config, volume_mountpoint[volume_name])
                else:
                    raise RuntimeError(
                        "Only 'host_type', 'temporary_directory' and 'persistent_volume' volumes are "
                        " supported by Docker executor,"
                        f"requested '{volume['config']['type']}' for {volume_name}."
                    )

        assert (
            constants.PROCESSING_VOLUME_NAME in results
        ), "Processing volume must be defined."
        return results

    def _init_volumes(self) -> Dict:
        """Prepare volumes for init container."""
        mount_points = [
            (config, mount_point, False)
            for config, mount_point in self._get_volumes().values()
        ]
        mount_points += [
            (connector.config, Path("/") / f"{storage_name}_{connector.name}", False)
            for storage_name, connector in get_mountable_connectors()
        ]
        return dict([self._new_volume(*mount_point) for mount_point in mount_points])

    def _communicator_volumes(self) -> Dict[str, Dict]:
        """Prepare volumes for communicator container."""
        mount_points = [
            (connector.config, Path("/") / f"{storage_name}_{connector.name}", False)
            for storage_name, connector in get_mountable_connectors()
        ]
        volumes = self._get_volumes()
        mount_points += [
            (*volumes[constants.SECRETS_VOLUME_NAME], False),
            (*volumes[constants.SOCKETS_VOLUME_NAME], False),
        ]
        return dict([self._new_volume(*mount_point) for mount_point in mount_points])

    def _processing_volumes(self) -> Dict:
        """Prepare volumes for processing container."""
        # Expose processing and (possibly) input volume RW.
        mount_points = [
            (config, mount_point, False)
            for config, mount_point in self._get_volumes(True).values()
        ]
        # Expose mountable connectors ('upload' RW, others 'RO').
        mount_points += [
            (
                connector.config,
                Path("/") / f"{storage_name}_{connector.name}",
                storage_name != "upload",
            )
            for storage_name, connector in get_mountable_connectors()
        ]

        mount_points += [
            (
                {"path": self.runtime_dir / "executors" / ExecutorFiles.SOCKET_UTILS},
                Path("/socket_utils.py"),
                False,
            ),
            (
                {"path": self.runtime_dir / constants.BOOTSTRAP_PYTHON_RUNTIME},
                Path("/") / constants.BOOTSTRAP_PYTHON_RUNTIME,
                False,
            ),
            (
                {"path": self.runtime_dir / "communicator.py"},
                Path("/communicator.py"),
                False,
            ),
            (
                {
                    "path": self.runtime_dir
                    / "executors"
                    / ExecutorFiles.STARTUP_PROCESSING_SCRIPT
                },
                Path("/start.py"),
                False,
            ),
            (
                {"path": self.runtime_dir / "executors" / ExecutorFiles.CONSTANTS},
                Path("/constants.py"),
                True,
            ),
        ]
        # Generate dummy passwd and create mappings for it. This is required because some tools
        # inside the container may try to lookup the given UID/GID and will crash if they don't
        # exist. So we create minimal user/group files.

        temporary_directory = Path(self.tmpdir.name)
        passwd_path = temporary_directory / "passwd"
        group_path = temporary_directory / "group"

        with passwd_path.open("wt") as passwd_file:
            passwd_file.write(
                "root:x:0:0:root:/root:/bin/bash\n"
                + f"user:x:{os.getuid()}:{os.getgid()}:user:{os.fspath(constants.PROCESSING_VOLUME)}:/bin/bash\n"
            )
        with group_path.open("wt") as group_file:
            group_file.write("root:x:0:\n" + f"user:x:{os.getgid()}:user\n")

        mount_points += [
            ({"path": passwd_path}, Path("/etc/passwd"), True),
            ({"path": group_path}, Path("/etc/group"), True),
        ]

        # Create mount points for tools.
        mount_points += [
            ({"path": Path(tool)}, self.tools_paths_prefix / str(index), True)
            for index, tool in enumerate(self.get_tools_paths())
        ]
        return dict([self._new_volume(*mount_point) for mount_point in mount_points])

    def _map_docker_image(self, docker_image: str) -> str:
        """Transform the given Docker image according to the mapper.

        If image starts with the prefix in the mapper keys, it is replaced with the
        corresponding value. If no prefix matches, the image is returned unchanged.
        """
        for prefix, replacement in SETTINGS.get("FLOW_CONTAINER_IMAGE_MAP", {}).items():
            if docker_image.startswith(prefix):
                return replacement + docker_image[len(prefix) :]
        return docker_image

    async def start(self):
        """Start process execution."""
        memory = (
            self.process["resource_limits"]["memory"] + DOCKER_MEMORY_HARD_LIMIT_BUFFER
        )
        memory_swap = int(memory * DOCKER_MEMORY_SWAP_RATIO)

        # By default the communication container is connected to the bridge network.
        network = SETTINGS.get("FLOW_EXECUTOR", {}).get("NETWORK", "bridge")

        # The network in the processing container depends on the process requirements.
        if "network" in self.resources:
            network_mode = f"container:{self.container_name}-communicator"
        else:
            network_mode = "none"

        processing_image = self._map_docker_image(
            self.requirements.get(
                "image",
                SETTINGS.get(
                    "FLOW_DOCKER_DEFAULT_PROCESSING_CONTAINER_IMAGE",
                    "public.ecr.aws/s4q6j6e8/resolwe/base:ubuntu-20.04",
                ),
            )
        )
        communicator_image = self._map_docker_image(
            SETTINGS.get(
                "FLOW_DOCKER_COMMUNICATOR_IMAGE",
                "public.ecr.aws/s4q6j6e8/resolwe/com:latest",
            )
        )
        ulimits = []
        if (
            self.process["scheduling_class"]
            == PROCESS_META["SCHEDULING_CLASS_INTERACTIVE"]
        ):
            # TODO: This is not very good as each child gets the same limit.
            # Note: Ulimit does not work as expected on multithreaded processes
            # Limit is increased by factor 1.2 for processes with 2-8 threads.
            # TODO: This should be changed for processes with over 8 threads.
            cpu_time_interactive = SETTINGS.get(
                "FLOW_PROCESS_RESOURCE_DEFAULTS", {}
            ).get("cpu_time_interactive", 30)
            cpu_limit = int(cpu_time_interactive * 1.2)
            ulimits.append(
                docker.types.Ulimit(name="cpu", soft=cpu_limit, hard=cpu_limit)
            )

        environment = {
            "LISTENER_PUBLIC_KEY": LISTENER_PUBLIC_KEY,
            "CURVE_PUBLIC_KEY": PUBLIC_KEY,
            "CURVE_PRIVATE_KEY": PRIVATE_KEY,
            "LISTENER_SERVICE_HOST": self.listener_connection[0],
            "LISTENER_SERVICE_PORT": self.listener_connection[1],
            "LISTENER_PROTOCOL": self.listener_connection[2],
            "DATA_ID": self.data_id,
            "RUNNING_IN_CONTAINER": 1,
            "RUNNING_IN_DOCKER": 1,
            "GENIALIS_UID": os.getuid(),
            "GENIALIS_GID": os.getgid(),
            "FLOW_MANAGER_KEEP_DATA": SETTINGS.get("FLOW_MANAGER_KEEP_DATA", False),
            "DESCRIPTOR_CHUNK_SIZE": 100,
            "MOUNTED_CONNECTORS": ",".join(
                connector.name
                for connector in connectors.values()
                if connector.mountable
            ),
            "TOOLS_PATHS": ":".join(
                f"{self.tools_paths_prefix / str(i)}"
                for i in range(len(self.get_tools_paths()))
            ),
        }
        with suppress(RuntimeError):
            environment["UPLOAD_DIR"] = get_upload_dir()

        autoremove = SETTINGS.get("FLOW_DOCKER_AUTOREMOVE", False)

        init_container_name = f"{self.container_name}-init"

        init_arguments = {
            "auto_remove": autoremove,
            "volumes": self._init_volumes(),
            "command": ["/usr/local/bin/python3", "-m", "executors.init_container"],
            "image": communicator_image,
            "name": init_container_name,
            "detach": True,
            "cpu_quota": 1000000,
            "mem_limit": "4000m",
            "mem_reservation": "200m",
            "network": network,
            # Init container must run as root to use named volumes.
            "user": "0:0",
            "environment": environment,
        }
        communication_arguments = {
            "auto_remove": autoremove,
            "volumes": self._communicator_volumes(),
            "command": ["/usr/local/bin/python", "/startup.py"],
            "image": communicator_image,
            "name": f"{self.container_name}-communicator",
            "detach": True,
            "cpu_quota": 100000,
            "mem_limit": "4000m",
            "mem_reservation": "200m",
            "network": network,
            "cap_drop": ["all"],
            "user": f"{os.getuid()}:{os.getgid()}",
            "environment": environment,
        }
        processing_arguments = {
            "auto_remove": autoremove,
            "volumes": self._processing_volumes(),
            "command": ["python3", "/start.py"],
            "image": processing_image,
            "network_mode": network_mode,
            "working_dir": os.fspath(self.processing_working_dir),
            "detach": True,
            "cpu_quota": self.process["resource_limits"]["cores"] * (10**6),
            "mem_limit": f"{memory}m",
            "mem_reservation": f"{self.process['resource_limits']['memory']}m",
            "mem_swappiness": DOCKER_MEMORY_SWAPPINESS,
            "memswap_limit": f"{memory_swap}m",
            "name": self.container_name,
            "cap_drop": ["all"],
            "user": f"{os.getuid()}:{os.getgid()}",
            "ulimits": ulimits,
            "environment": environment,
        }

        @retry(max_retries=5)
        def transfer_image(client, image_name):
            """Transfer missing image, retry 5 times."""
            client.images.pull(image_name)

        client = docker.from_env()
        # Pull all the images.
        try:
            try:
                logger.debug("Pulling processing image %s.", processing_image)
                client.images.get(processing_image)
            except docker.errors.ImageNotFound:
                transfer_image(client, processing_image)
            try:
                logger.debug("Pulling communicator image %s.", communicator_image)
                client.images.get(communicator_image)
            except docker.errors.ImageNotFound:
                transfer_image(client, communicator_image)

        except docker.errors.APIError:
            logger.exception("Docker API error")
            raise RuntimeError("Docker API error")

        loop = asyncio.get_event_loop()
        start_time = time.time()
        try:
            init_container = client.containers.run(**init_arguments)
        except docker.errors.APIError as error:
            await self.communicator.finish(
                {"error": f"Error starting init container: {error}"}
            )
            raise

        init_container_status = await loop.run_in_executor(None, init_container.wait)

        # Return code is as follows:
        # - 0: no error occured, continue processing.
        # - 1: error running init container, abort processing and log error.
        # - 2: data exists in the processing volume, abort processing.
        init_rc = init_container_status["StatusCode"]
        if init_rc != 0:
            logger.error("Init container returned %s instead of 0.", init_rc)
            # Do not set error on data objects where previous data exists.
            if init_rc == 1:
                await self.communicator.finish(
                    {"error": f"Init container returned {init_rc} instead of 0."}
                )
            return

        try:
            communication_container = client.containers.run(**communication_arguments)
        except docker.errors.APIError as error:
            await self.communicator.finish(
                {"error": f"Error starting communication container: {error}"}
            )
            raise
        try:
            processing_container = client.containers.run(**processing_arguments)
        except docker.errors.APIError as e:
            await self.communicator.finish(
                {"error": f"Error starting processing container: {e}"}
            )
            with suppress(docker.errors.APIError):
                communication_container.stop(timeout=1)
            raise

        end_time = time.time()
        logger.info(
            "It took {:.2f}s for Docker containers to start".format(
                end_time - start_time
            )
        )
        with suppress(docker.errors.NotFound):
            await loop.run_in_executor(None, communication_container.wait)
        with suppress(docker.errors.NotFound):
            await loop.run_in_executor(None, processing_container.wait)



"""Background task viewset."""

from django_filters import rest_framework as filters
from rest_framework import mixins, serializers, viewsets

from resolwe.observers.mixins import ObservableMixin
from resolwe.observers.models import BackgroundTask

STATUSES = [status[0] for status in BackgroundTask.STATUS_CHOICES]


class BackgroundTaskFilter(filters.FilterSet):
    """Filter for BackgroundTask."""

    class Meta:
        """Set the filter lookups for the status field."""

        model = BackgroundTask
        fields = {"status": ["exact", "in"], "id": ["exact", "in"]}


class BackgroundTaskSerializer(serializers.ModelSerializer):
    """Serialize the BackgroundTask model."""

    class Meta:
        """Set the model."""

        model = BackgroundTask
        fields = ("id", "started", "finished", "status", "description", "output")


class BackgroundTaksViewSet(
    ObservableMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    viewsets.GenericViewSet,
):
    """Observable task viewset.

    It exposes standard observable endpoints so user can subscribe to it plus it lists
    all the background tasks for a current user.
    """

    serializer_class = BackgroundTaskSerializer
    queryset = BackgroundTask.objects.all()
    filterset_class = BackgroundTaskFilter



# Generated by Django 4.2.2 on 2023-10-08 20:19

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0015_alter_annotationfield_vocabulary"),
    ]

    operations = [
        migrations.AddConstraint(
            model_name="annotationvalue",
            constraint=models.UniqueConstraint(
                fields=("entity", "field"), name="uniquetogether_entity_field"
            ),
        ),
    ]



# Generated by Django 4.2.7 on 2024-01-09 07:05

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0020_alter_annotationvalue_field"),
    ]

    operations = [
        migrations.AddField(
            model_name="annotationvalue",
            name="modified",
            field=models.DateTimeField(auto_now=True),
        ),
    ]



"""Base model for all Resolwe models."""

import datetime
from enum import StrEnum
from typing import TypeVar

from django.conf import settings
from django.db import IntegrityError, models, transaction
from versionfield import VersionField

from resolwe.auditlog.models import AuditModel
from resolwe.flow.models.fields import ResolweSlugField
from resolwe.permissions.models import PermissionManager, PermissionQuerySet

VERSION_NUMBER_BITS = (8, 10, 14)

# Maximum number of slug-generation retries.
MAX_SLUG_RETRIES = 10
MAX_SLUG_LENGTH = 100
MAX_NAME_LENGTH = 100

# Data status enum.


class DataStatus(StrEnum):
    """Status of the data object."""

    #: data object is uploading
    UPLOADING = "UP"
    #: data object is being resolved
    RESOLVING = "RE"
    #: data object is waiting
    WAITING = "WT"
    #: data object is preparing
    PREPARING = "PP"
    #: data object is processing
    PROCESSING = "PR"
    #: data object is done
    DONE = "OK"
    #: data object is in error state
    ERROR = "ER"
    #: data object is in dirty state
    DIRTY = "DR"

    @classmethod
    def sort_order(cls):
        """Return a priority for sorting data statuses in a collection."""
        return [
            cls.ERROR,
            cls.PROCESSING,
            cls.UPLOADING,
            cls.PREPARING,
            cls.WAITING,
            cls.RESOLVING,
            cls.DONE,
        ]


M = TypeVar("M", bound=models.Model)
Q = TypeVar("Q", bound=models.QuerySet)


# NOTE: This method is used in migrations.
def delete_chunked(queryset, chunk_size=500):
    """Chunked delete, which should be used if deleting many objects.

    The reason why this method is needed is that deleting a lot of Data objects
    requires Django to fetch all of them into memory (fast path is not used) and
    this causes huge memory usage (and possibly OOM).

    :param chunk_size: Optional chunk size
    """
    while True:
        # Discover primary key to limit the current chunk. This is required because delete
        # cannot be called on a sliced queryset due to ordering requirement.
        with transaction.atomic():
            # Get offset of last item (needed because it may be less than the chunk size).
            offset = queryset.order_by("pk")[:chunk_size].count()
            if not offset:
                break

            # Fetch primary key of last item and use it to delete the chunk.
            last_instance = queryset.order_by("pk")[offset - 1]
            queryset.filter(pk__lte=last_instance.pk).delete()


class BaseQuerySet(PermissionQuerySet):
    """Base query set for Resolwe's ORM objects."""

    def delete_chunked(self, chunk_size=500):
        """Chunked delete, which should be used if deleting many objects.

        The reason why this method is needed is that deleting a lot of Data objects
        requires Django to fetch all of them into memory (fast path is not used) and
        this causes huge memory usage (and possibly OOM).

        :param chunk_size: Optional chunk size
        """
        return delete_chunked(self, chunk_size=chunk_size)


class BaseManager(PermissionManager[M, Q]):
    """Base manager set for Resolwe's ORM objects."""

    # The data is partitioned into groups with the same partition fields value. The
    # object with the greatest version inside the partition is considered the latest.
    partition_fields: list[str] = ["slug"]

    # The field to use as version. The latest version is determined by sorting the
    # data by version field in the ascending order and taking the first entry.
    version_field: str = "version"

    # The created field is used to determine the version at certain point in time.
    created_field: str = "created"

    # Versioning is enabled by default.
    enable_versioning: bool = True

    # Use BaseQuerySet by default.
    QuerySet = BaseQuerySet

    def _only_latest(self, queryset: Q) -> Q:
        """Return only latest values from the queryset.

        The queryset is partitioned by the partition fields and the latest version from
        each partition is returned.
        """
        latest_entries = queryset.annotate(
            rank=models.Window(
                expression=models.functions.DenseRank(),
                partition_by=[models.F(field) for field in self.partition_fields],
                order_by=models.F(self.version_field).desc(),
            ),
        ).filter(rank=1)
        return latest_entries

    def get_queryset(self) -> Q:
        """Return only the latest version for every field.

        Make sure the queryset is lazily evaluated, since the get_queryset method can
        be called many times (for instance, when filtering the queryset).
        """
        queryset = super().get_queryset()
        return self._only_latest(queryset) if self.enable_versioning else queryset

    def history(self, timestamp: datetime.datetime) -> Q:
        """Get values at certain point in time."""
        created_filter = {f"{self.created_field}__le": timestamp}
        filtered_queryset: Q = self.filter(**created_filter)  # type: ignore
        return self._only_latest(filtered_queryset)


class BaseManagerWithoutVersion(BaseManager[M, Q]):
    """Base manager for Resolwe's ORM objects without versioning."""

    # Disable object versioning.
    enable_versioning: bool = False


class ModifiedField(models.DateTimeField):
    """Custom datetime field.

    When attribute 'skip_auto_now' is set to True, the field is not updated in the
    pre_save method even if auto_now is set to True. The 'skip_auto_now' is reset
    on every save to the default value 'False'.
    """

    def pre_save(self, model_instance, add):
        """Prepare field for saving."""
        # When skip_auto_now is set do not generate the new value even is auto_now is
        # set to True.
        if getattr(model_instance, "skip_auto_now", False) is True:
            return getattr(model_instance, self.attname)
        else:
            return super().pre_save(model_instance, add)


class UniqueSlugError(IntegrityError):
    """The error indicates slug collision."""


class BaseModel(AuditModel):
    """Abstract model that includes common fields for other models."""

    class Meta:
        """BaseModel Meta options."""

        abstract = True
        unique_together = ("slug", "version")
        default_permissions = ()
        get_latest_by = "version"

    #: URL slug
    slug = ResolweSlugField(
        populate_from="name", unique_with="version", max_length=MAX_SLUG_LENGTH
    )

    #: process version
    version = VersionField(number_bits=VERSION_NUMBER_BITS, default="0.0.0")

    #: object name
    name = models.CharField(max_length=MAX_NAME_LENGTH)

    #: creation date and time
    created = models.DateTimeField(auto_now_add=True, db_index=True)

    #: modified date and time
    modified = ModifiedField(auto_now=True, db_index=True)

    #: user that created the entry
    contributor = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT)

    #: the latest version of the objects
    objects = BaseManager()

    #: all abjects
    all_objects = models.Manager()

    def __str__(self):
        """Format model name."""
        return self.name

    def save(self, *args, **kwargs):
        """Save the model."""
        name_max_len = self._meta.get_field("name").max_length
        if len(self.name) > name_max_len:
            self.name = self.name[: (name_max_len - 3)] + "..."

        for _ in range(MAX_SLUG_RETRIES):
            try:
                # Attempt to save the model. It may fail due to slug conflict.
                with transaction.atomic():
                    super().save(*args, **kwargs)
                    break
            except IntegrityError as error:
                # Retry in case of slug conflicts.
                if "{}_slug".format(self._meta.db_table) in error.args[0]:
                    self.slug = None
                    continue

                raise
        else:
            raise UniqueSlugError(
                f"Unable to generate unique slug after {MAX_SLUG_RETRIES} retries."
            )



""".. Ignore pydocstyle D400.

========
Consumer
========

Manager Channels consumer.

"""

import asyncio
import logging
import os
from contextlib import suppress
from pathlib import Path

from channels.consumer import AsyncConsumer
from channels.db import database_sync_to_async
from channels.layers import get_channel_layer
from channels.testing import ApplicationCommunicator
from django.db import connection

from resolwe.utils import BraceMessage as __

from . import state

logger = logging.getLogger(__name__)
CHANNEL_HEALTH_CHECK = os.environ.get("HOSTNAME")


async def send_event(message):
    """Construct a Channels event packet with the given message.

    :param message: The message to send to the manager workers.
    """
    packet = {
        "type": "control_event",  # This is used as the method name in the consumer.
        "content": message,
    }
    await get_channel_layer().send(state.MANAGER_CONTROL_CHANNEL, packet)


async def run_consumer(timeout=None):
    """Run the consumer until it finishes processing.

    :param timeout: Set maximum execution time before cancellation, or
        ``None`` (default) for unlimited.
    """
    manager_channel = state.MANAGER_CONTROL_CHANNEL
    manager_scope = {
        "type": "control_event",
        "channel": manager_channel,
    }
    manager_app = ApplicationCommunicator(ManagerConsumer(), manager_scope)
    channel_layer = get_channel_layer()

    async def _consume_loop(channel, scope, app):
        """Run a loop to consume messages off the channels layer."""
        message = await channel_layer.receive(channel)
        while message.get("type", {}) != "_resolwe_manager_quit":
            message.update(scope)
            await app.send_input(message)
            message = await channel_layer.receive(channel)

    consume_future = asyncio.ensure_future(
        _consume_loop(manager_channel, manager_scope, manager_app)
    )

    with suppress(asyncio.TimeoutError):
        await asyncio.wait_for(consume_future, timeout=timeout)

    await manager_app.wait()


async def exit_consumer():
    """Cause the synchronous consumer to exit cleanly."""
    packet = {
        "type": "_resolwe_manager_quit",
    }
    await get_channel_layer().send(state.MANAGER_CONTROL_CHANNEL, packet)


class ManagerConsumer(AsyncConsumer):
    """Channels consumer for handling manager events."""

    def __init__(self, *args, **kwargs):
        """Initialize a consumer instance with the given manager."""
        # This import is local in order to avoid startup import loops.
        from . import manager

        self.manager = manager

    async def control_event(self, message):
        """Forward control events to the manager dispatcher."""
        logger.debug("control_event got message %s", message)
        try:
            await self.manager.handle_control_event(message["content"])
        except:
            logger.exception("control_event exception.")


class HealtCheckConsumer(AsyncConsumer):
    """Channels consumer for handling health-check events."""

    async def health_check(self, message: dict):
        """Perform health check.

        We are testing the channels layer and database layer. The channels
        layer is already functioning if this method is called so we have to
        perform database check.

        If the check is successfull touch the file specified in the channels
        message.
        """
        logger.debug(__("Performing health check with message {}.", message))
        path = Path(message["file"])

        if await database_sync_to_async(self.check_database, thread_sensitive=False)():
            logger.debug("Health check passed.")
            path.touch(exist_ok=True)

    def check_database(self) -> bool:
        """Perform a simple database check."""
        with connection.cursor() as cursor:
            cursor.execute("SELECT 1;")
            result = cursor.fetchone()[0]
            return result == 1



"""Define the class to handle exposing models to the python processes."""

from typing import TYPE_CHECKING, Type

from django.contrib.auth import get_user_model
from django.db.models import Model, QuerySet

from resolwe.flow.managers.listener.plugin_interface import Plugin, PluginManager
from resolwe.flow.models import (
    Collection,
    Data,
    DescriptorSchema,
    Entity,
    Process,
    Storage,
)
from resolwe.permissions.models import Permission

if TYPE_CHECKING:
    from django.contrib.auth.models import User as UserType

User = get_user_model()


class PermissionManager(PluginManager["ExposeObjectPlugin"]):
    """Permission manager class."""

    plugin_unique_identifier = "full_model_name"

    def can_handle(self, full_model_name: str) -> bool:
        """Query permission manager if it can handle given model."""
        return full_model_name in self._plugins

    def can_create(self, user, full_model_name: str, attributes, data_id: int):
        """Query if user can create the given model.

        :raises RuntimeError: if user does not have permissions to create the
            given model.
        :raises KeyError: if permission manager has no plugin registered for
            the given model.
        """
        return self._plugins[full_model_name].can_create(user, attributes, data_id)

    def can_update(
        self,
        user,
        full_model_name: str,
        model_instance: Model,
        attributes,
        data_id: int,
    ):
        """Query if user can update the given model.

        :raises RuntimeError: if user does not have permissions to update the
            given model.
        :raises KeyError: if permission manager has no plugin registered for
            the given model.
        """
        return self._plugins[full_model_name].can_update(
            user, model_instance, attributes, data_id
        )

    def filter_objects(
        self, user, full_model_name: str, queryset: QuerySet, data_id: int
    ) -> QuerySet:
        """Filter the objects for the given user.

        :raises KeyError: if permission manager has no plugin registered for
            the given model.
        """
        return self._plugins[full_model_name].filter_objects(user, queryset, data_id)

    def can_read(self, user, full_model_name: str, data_id: int):
        """Query if user can read the metadata of the given model.

        :raises RuntimeError: if user does not have permissions to read the
            given model.
        :raises KeyError: if permission manager has no plugin registered for
            the given model.
        """
        return self._plugins[full_model_name].can_read(user, data_id)


permission_manager = PermissionManager()


class ExposeObjectPlugin(Plugin):
    """Plugin for exposing models in Python processes."""

    full_model_name = "app_label.model_name"
    plugin_manager = permission_manager
    abstract = True

    def _has_permission(
        self, user: "UserType", model: Type[Model], model_pk: int, permission_name: str
    ):
        """Check if contributor has requested permissions.

        :raises RuntimeError: with detailed explanation when check fails.
        """
        object_ = model.objects.filter(pk=model_pk)
        if not object_.exists():
            raise RuntimeError(
                f"Object {model._meta.model_name} with id {model_pk} not found."
            )

        if not object_.filter_for_user(
            user, Permission.from_name(permission_name)
        ).exists():
            if object_:
                raise RuntimeError(
                    f"No edit permission for {model._meta.model_name} with id {model_pk}."
                )
            else:
                raise RuntimeError(
                    f"Object {model._meta.model_name} with id {model_pk} not found."
                )

    def can_create(self, user: "UserType", attributes: dict, data_id: int):
        """Can user create the model with given attributes.

        :raises RuntimeError: if user does not have permissions to create the
            given model.
        """
        raise RuntimeError(
            (
                f"User {user} has no permission to create model"
                f"{self.full_model_name} with attributes {attributes}."
            )
        )

    def can_update(
        self, user: "UserType", model_instance: Model, attributes: dict, data_id: int
    ):
        """Can user update the given model instance.

        :raises RuntimeError: when user does not have permissions to update
            the given model.
        """
        raise RuntimeError(
            (
                f"User {user} has no permission to update model"
                f"{self.full_model_name} with attributes {attributes}."
            )
        )

    def filter_objects(
        self,
        user: "UserType",
        queryset: QuerySet,
        data_id: int,
    ) -> QuerySet:
        """Filter the objects for the given user."""
        return queryset.filter_for_user(user)

    def can_read(self, user: "UserType", data_id: int):
        """Can read model structural info.

        :raises RuntimeError: when user does not have permissions to access
            the given model.
        """

    @classmethod
    def get_identifier(cls):
        """Get the plugin identifier."""
        return cls.full_model_name


class ExposeData(ExposeObjectPlugin):
    """Expose the Data model."""

    full_model_name = "flow.Data"

    def can_create(self, user: "UserType", model_data: dict, data_id: int):
        """Can user update the given model instance.

        :raises RuntimeError: if user does not have permissions to create the
            given model.
        """
        allowed_fields = {
            "process_id",
            "output",
            "input",
            "tags",
            "entity",
            "entity_id",
            "collection",
            "collection_id",
            "name",
        }
        not_allowed_keys = set(model_data.keys()) - allowed_fields
        if not_allowed_keys:
            raise RuntimeError(f"Not allowed to set {','.join(not_allowed_keys)}.")

        # Check process permissions.
        self._has_permission(user, Process, model_data["process_id"], "view")

        if "entity_id" in model_data:
            # Check entity permissions.
            self._has_permission(user, Entity, model_data["entity_id"], "edit")

        if "collection_id" in model_data:
            # Check collection permissions.
            self._has_permission(user, Collection, model_data["collection_id"], "edit")

    def can_update(
        self, user: "UserType", model_instance: Data, model_data: dict, data_id: int
    ):
        """Can user update the given model instance.

        :raises RuntimeError: when user does not have permissions to update
            the given model.
        """
        allowed_fields = {
            "output",
            "tags",
            "entity_id",
            "entity",
            "collection_id",
            "collection",
            "name",
            "descriptor_schema_id",
            "descriptor_schema",
            "descriptor",
        }
        not_allowed_keys = set(model_data.keys()) - allowed_fields
        if not_allowed_keys:
            raise RuntimeError(f"Not allowed to set {','.join(not_allowed_keys)}.")

        # Check permission to modify the Data object. The current data object
        # can always be modified.
        if model_instance.id != data_id:
            self._has_permission(user, Data, model_instance.id, "edit")

        if "entity_id" in model_data:
            # Check entity permissions.
            self._has_permission(user, Entity, model_data["entity_id"], "edit")

        if "collection_id" in model_data:
            # Check collection permissions.
            self._has_permission(user, Collection, model_data["collection_id"], "edit")
        if "descriptor_schema_id" in model_data:
            # Check DescriptorSchema permissions.
            self._has_permission(
                user, DescriptorSchema, model_data["descriptor_schema_id"], "view"
            )

    def filter_objects(
        self,
        user: "UserType",
        queryset: QuerySet,
        data_id: int,
    ) -> QuerySet:
        """Filter the objects for the given user."""
        parent_ids = Data.objects.filter(pk=data_id).values_list("parents")
        inputs = queryset.filter(id__in=parent_ids)
        return queryset.filter_for_user(user).distinct().union(inputs)


class ExposeDescriptorSchema(ExposeObjectPlugin):
    """Expose the DescriptorSchema model."""

    full_model_name = "flow.DescriptorSchema"


class ExposeUser(ExposeObjectPlugin):
    """Expose the User model.

    Used to read contributor data.
    """

    full_model_name = User._meta.label

    def filter_objects(
        self,
        user: "UserType",
        queryset: QuerySet,
        data_id: int,
    ) -> QuerySet:
        """Filter the objects for the given user."""
        return queryset.filter(pk=user.id)


class ExposeEntity(ExposeObjectPlugin):
    """Expose the Entity model."""

    full_model_name = "flow.Entity"

    def can_create(self, user: "UserType", attributes: dict, data_id: int):
        """Can user update the given model instance.

        :raises RuntimeError: if user does not have permissions to create the
            given model.
        """

    def can_update(
        self, user: "UserType", model_instance: Data, model_data: dict, data_id: int
    ):
        """Can user update the given model instance.

        :raises RuntimeError: when user does not have permissions to update
            the given model.
        """
        allowed_fields = {
            "description",
            "tags",
            "name",
            "descriptor",
        }
        not_allowed_keys = set(model_data.keys()) - allowed_fields
        if not_allowed_keys:
            raise RuntimeError(f"Not allowed to set {','.join(not_allowed_keys)}.")

        # Check permission to modify the Entity object.
        self._has_permission(user, Entity, model_instance.id, "edit")


class ExposeCollection(ExposeObjectPlugin):
    """Expose the Collection model."""

    full_model_name = "flow.Collection"

    def can_create(self, user: "UserType", attributes: dict, data_id: int):
        """Can user update the given model instance.

        :raises RuntimeError: if user does not have permissions to create the
            given model.
        """


class ExposeProcess(ExposeObjectPlugin):
    """Expose the Process model."""

    full_model_name = "flow.Process"

    def filter_objects(
        self,
        user: "UserType",
        queryset: QuerySet,
        data_id: int,
    ) -> QuerySet:
        """Filter the objects for the given user."""
        parent_ids = Data.objects.filter(pk=data_id).values_list("parents")
        processes_of_inputs = queryset.filter(data__in=parent_ids).distinct()
        return queryset.filter_for_user(user).distinct().union(processes_of_inputs)


class ExposeStorage(ExposeObjectPlugin):
    """Expose the Storage model."""

    full_model_name = "flow.Storage"

    def can_create(self, user: "UserType", attributes: dict, data_id: int):
        """Can user create the model with given attributes.

        :raises RuntimeError: when user does not have permissions to create
            the given model.
        """

    def can_update(
        self, user: "UserType", model_instance: Storage, model_data: dict, data_id: int
    ):
        """Can user update the given Storage object.

        :raises RuntimeError: when user does not have permissions to update
            the given model.
        """
        processed_data_ids = set()
        # User must have permission to modify all the data objects this object belongs to.
        for data in model_instance.data.all():
            permission_manager.can_update(
                user, "flow.Data", data, {"output": {}}, data_id
            )
            processed_data_ids.add(data_id)

        # Only contributor can modify Storage object if it is orphaned.
        if not processed_data_ids and model_instance.contributor != user:
            raise RuntimeError(
                "Only contributor can modify unassigned Storage object with "
                f"id {model_instance.pk}, modification attempted as {user}."
            )

        # When adding storage to Data objects check permissions to modify the
        # data objects. Since permission checks are slow only process the data
        # objects that were not processed above.
        for data_id in model_data.get("data", []):
            if data_id not in processed_data_ids:
                permission_manager.can_update(
                    user,
                    "flow.Data",
                    Data.objects.get(pk=data_id),
                    {"output": {}},
                    data_id,
                )

    def filter_objects(
        self,
        user: "UserType",
        queryset: QuerySet,
        data_id: int,
    ) -> QuerySet:
        """Filter the objects for the given user."""
        return (
            queryset.filter_for_user(user)
            .distinct()
            .union(queryset.filter(contributor=user))
        )



# pylint: disable=missing-docstring
from django.apps import apps
from django.conf import settings
from django.core.management import call_command
from django.db.utils import DataError
from django.test import TestCase, override_settings

from resolwe.flow.exceptions import SlugError

CUSTOM_SETTINGS = {
    "INSTALLED_APPS": settings.INSTALLED_APPS + ("resolwe.flow.tests.fields_test_app",),
}


@override_settings(**CUSTOM_SETTINGS)
class FieldsTest(TestCase):
    def setUp(self):
        super().setUp()

        apps.clear_cache()
        call_command("migrate", verbosity=0, interactive=False)
        # NOTE: Models must be imported after migrations are run.

    def test_populate_from_name(self):
        from .fields_test_app.models import TestModel

        obj = TestModel.objects.create(name="Test object")
        self.assertEqual(obj.slug, "test-object")

        obj = TestModel.objects.create(name="Test object")
        self.assertEqual(obj.slug, "test-object-2")

        # Base of the slug shouldn't be changed even if it finishes with number.
        obj = TestModel.objects.create(name="Test object 2")
        self.assertEqual(obj.slug, "test-object-2-2")

    def test_predefined_slug(self):
        from .fields_test_app.models import TestModel

        obj = TestModel.objects.create(name="Test object")

        # It is ok to add a sequence to slug if it is not explicitly defined.
        TestModel.objects.create(name="Test object")

        with self.assertRaisesRegex(SlugError, "is already taken"):
            TestModel.objects.create(name="Test object", slug=obj.slug)

        # Predefined slug must NOT be changed.
        with self.assertRaisesRegex(SlugError, "not a valid slug"):
            TestModel.objects.create(name="Test object", slug="Not a valid slug")

    def test_predefined_long_slugs(self):
        from .fields_test_app.models import TestModel

        slug_max_length = TestModel._meta.get_field("slug").max_length
        max_slug = "x" * slug_max_length

        # Predefined slug shouldn't be trimmed
        obj = TestModel.objects.create(name="Test object", slug=max_slug)
        self.assertEqual(obj.slug, max_slug)

        with self.assertRaisesRegex(DataError, "value too long"):
            TestModel.objects.create(name="Test object", slug=max_slug + "y")

    def test_generate_only_on_create(self):
        from .fields_test_app.models import TestModel

        obj = TestModel.objects.create(name="Test object")
        obj.save()  # Second save shouldn't add sequence number to slug.
        self.assertEqual(obj.slug, "test-object")

        obj.name = "Different name"
        obj.save()
        self.assertEqual(obj.slug, "test-object")

    def test_unique_with(self):
        from .fields_test_app.models import TestModel

        obj = TestModel.objects.create(name="Test object", version="1.0.0")
        self.assertEqual(obj.slug, "test-object")

        obj = TestModel.objects.create(name="Test object", version="2.0.0")
        self.assertEqual(obj.slug, "test-object")

        obj.version = "1.0.0"
        with self.assertRaisesRegex(SlugError, "is already taken"):
            obj.save()

    def test_long_sequence(self):
        from .fields_test_app.models import TestModel

        TestModel.objects.create(name="Test object", slug="test-object")
        TestModel.objects.create(name="Test object", slug="test-object-999999999")

        with self.assertRaisesRegex(SlugError, "slug sequence too long"):
            TestModel.objects.create(name="Test object")

    def test_empty_name(self):
        from .fields_test_app.models import TestModel

        obj = TestModel.objects.create()
        self.assertEqual(obj.slug, "testmodel")

        # Empty slugified name.
        obj = TestModel.objects.create(name="?")
        self.assertEqual(obj.slug, "testmodel-2")



# pylint: disable=missing-docstring
from django.contrib.auth import get_user_model
from django.test import TransactionTestCase as DjangoTransactionTestCase
from rest_framework import status

from resolwe.flow.models import Collection
from resolwe.flow.views import CollectionViewSet
from resolwe.observers.models import BackgroundTask
from resolwe.permissions.models import get_anonymous_user
from resolwe.test import ResolweAPITestCase, TransactionResolweAPITestCase

MESSAGES = {
    "NOT_FOUND": "No Collection matches the given query.",
    "NO_PERMISSION": "You do not have permission to perform this action.",
}


class CollectionTestCaseCommonMixin:
    fixtures = [
        "users.yaml",
        "permissions.yaml",
        "data.yaml",
        "collections.yaml",
        "processes.yaml",
    ]

    def setUp(self):
        self.collection1 = Collection.objects.get(pk=1)

        self.resource_name = "collection"
        self.viewset = CollectionViewSet

        self.post_data = {
            "name": "Test collection",
            "slug": "test_collection",
        }

        # Reset the anonymous user for each test. This is important since permissions
        # are set in fixtures but id of the public user changes between tests.
        get_anonymous_user(cache=False)
        super().setUp()


class CollectionTestCase(CollectionTestCaseCommonMixin, ResolweAPITestCase):
    def test_get_list(self):
        resp = self._get_list(self.user1)
        self.assertEqual(len(resp.data), 3)

    def test_get_list_public_user(self):
        # public user
        resp = self._get_list()
        self.assertEqual(len(resp.data), 2)
        # check that you get right two objects
        self.assertEqual([p["id"] for p in resp.data], [1, 3])
        # check that (one of the) objects have expected keys
        self.assertKeys(
            resp.data[0],
            [
                "slug",
                "status",
                "name",
                "created",
                "modified",
                "contributor",
                "description",
                "id",
                "settings",
                "current_user_permissions",
                "data_count",
                "descriptor_schema",
                "descriptor",
                "descriptor_dirty",
                "tags",
                "duplicated",
                "entity_count",
            ],
        )

    def test_get_list_admin(self):
        resp = self._get_list(self.admin)
        self.assertEqual(len(resp.data), 3)

    def test_get_list_groups(self):
        resp = self._get_list(self.user3)
        self.assertEqual(len(resp.data), 3)

    def test_post(self):
        # logged-in user
        collection_n = Collection.objects.count()
        resp = self._post(self.post_data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_201_CREATED)
        self.assertEqual(Collection.objects.count(), collection_n + 1)

        # public user
        collection_n += 1
        resp = self._post(self.post_data)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(Collection.objects.count(), collection_n)

    def test_get_detail(self):
        # public user w/ perm
        resp = self._get_detail(1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertEqual(resp.data["id"], 1)

        # public user w/o perm
        resp = self._get_detail(2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(resp.data["detail"], MESSAGES["NOT_FOUND"])

        # user w/ permissions
        resp = self._get_detail(1, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertKeys(
            resp.data,
            [
                "slug",
                "status",
                "name",
                "created",
                "modified",
                "contributor",
                "description",
                "settings",
                "id",
                "current_user_permissions",
                "data_count",
                "descriptor_schema",
                "descriptor",
                "descriptor_dirty",
                "tags",
                "duplicated",
                "entity_count",
            ],
        )

        # user w/o permissions
        resp = self._get_detail(2, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(resp.data["detail"], MESSAGES["NOT_FOUND"])

        # user w/ public permissions
        resp = self._get_detail(1, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        self.assertKeys(
            resp.data,
            [
                "slug",
                "status",
                "name",
                "created",
                "modified",
                "contributor",
                "description",
                "settings",
                "id",
                "current_user_permissions",
                "data_count",
                "descriptor_schema",
                "descriptor",
                "descriptor_dirty",
                "tags",
                "duplicated",
                "entity_count",
            ],
        )

    def test_patch(self):
        data = {"name": "New collection"}
        resp = self._patch(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_200_OK)
        p = Collection.objects.get(pk=1)
        self.assertEqual(p.name, "New collection")

        # protected field
        data = {"created": "3042-01-01T09:00:00"}
        resp = self._patch(1, data, self.user1)
        self.assertEqual(resp.status_code, status.HTTP_204_NO_CONTENT)
        p = Collection.objects.get(pk=1)
        self.assertEqual(p.created.isoformat(), self.collection1.created.isoformat())

    def test_patch_no_perm(self):
        data = {"name": "New collection"}
        resp = self._patch(2, data, self.user2)
        self.assertEqual(resp.status_code, status.HTTP_404_NOT_FOUND)
        p = Collection.objects.get(pk=2)
        self.assertEqual(p.name, "Test collection 2")

    def test_patch_public_user(self):
        data = {"name": "New collection"}
        resp = self._patch(3, data)
        self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)
        p = Collection.objects.get(pk=3)
        self.assertEqual(p.name, "Test collection 3")


class CollectionTestCaseDelete(
    CollectionTestCaseCommonMixin,
    TransactionResolweAPITestCase,
    DjangoTransactionTestCase,
):
    """Test deleting collections.

    We have to inherit from DjangoTransactionTestCase because the deletion in done in a
    background task. The tests are in a separate class not to slow down the other tests.
    """

    def _pre_setup(self, *args, **kwargs):
        """Delete all previously created users.

        The public user is created in the migrations and recreated during test case,
        but with a different id which breaks the loading of the fixtures.
        """
        get_user_model().objects.all().delete()
        super()._pre_setup(*args, **kwargs)

    def test_delete(self):
        response = self._delete(1, self.user1)
        BackgroundTask.objects.get(pk=response.data["id"]).wait()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        collection_exists = Collection.objects.filter(pk=1).exists()
        self.assertFalse(collection_exists)

    def test_delete_no_perm(self):
        response = self._delete(2, self.user2)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        collection_exists = Collection.objects.filter(pk=2).exists()
        self.assertTrue(collection_exists)

    def test_delete_public_user(self):
        response = self._delete(3)
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        collection_exists = Collection.objects.filter(pk=3).exists()
        self.assertTrue(collection_exists)



""".. Ignore pydocstyle D400.

==========================
Resolwe Storage Connectors
==========================

"""

from .baseconnector import DEFAULT_CONNECTOR_PRIORITY, ConnectorType
from .localconnector import LocalFilesystemConnector
from .registry import StorageConnectors, connectors
from .transfer import Transfer

try:
    from .s3connector import AwsS3Connector
except ImportError:
    AwsS3Connector = None
try:
    from .googleconnector import GoogleConnector
except ImportError:
    GoogleConnector = None


__all__ = (
    "AwsS3Connector",
    "GoogleConnector",
    "LocalFilesystemConnector",
    "StorageConnectors",
    "Transfer",
    "DEFAULT_CONNECTOR_PRIORITY",
    "connectors",
    "ConnectorType",
)



# Generated by Django 4.2.9 on 2024-02-13 07:00

from django.contrib.postgres.operations import UnaccentExtension
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0024_update_sql_entity_methods"),
    ]

    operations = [UnaccentExtension()]



# Generated by Django 4.2.16 on 2024-10-09 08:50

import versionfield.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0028_update_all_sql_entity_methods"),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name="annotationfield",
            name="uniquetogether_name_group",
        ),
        migrations.AddField(
            model_name="annotationfield",
            name="version",
            field=versionfield.fields.VersionField(default="0.0.0"),
        ),
        migrations.AddConstraint(
            model_name="annotationfield",
            constraint=models.UniqueConstraint(
                fields=("name", "group", "version"),
                name="uniquetogether_name_group_version",
            ),
        ),
    ]



# Generated by Django 3.1.7 on 2021-10-12 11:34

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="PermissionGroup",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="PermissionModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("value", models.PositiveSmallIntegerField()),
                (
                    "group",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="model_permissions",
                        to="auth.group",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="model_permissions",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "permission_group",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="permissions",
                        to="permissions.permissiongroup",
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="permissionmodel",
            constraint=models.UniqueConstraint(
                condition=models.Q(user__isnull=False),
                fields=("permission_group", "value", "user"),
                name="one_permission_per_user",
            ),
        ),
        migrations.AddConstraint(
            model_name="permissionmodel",
            constraint=models.UniqueConstraint(
                condition=models.Q(group__isnull=False),
                fields=("permission_group", "value", "group"),
                name="one_permission_per_group",
            ),
        ),
        migrations.AddConstraint(
            model_name="permissionmodel",
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(("group__isnull", True), ("user__isnull", False)),
                    models.Q(("group__isnull", False), ("user__isnull", True)),
                    _connector="OR",
                ),
                name="exactly_one_of_user_group_must_be_set",
            ),
        ),
    ]



# Generated by Django 3.2.12 on 2022-11-21 08:36

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("permissions", "0004_create_anonymous_user"),
        ("flow", "0007_alter_data_process_rc"),
    ]

    operations = [
        migrations.CreateModel(
            name="AnnotationField",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128)),
                ("type", models.CharField(max_length=16)),
                ("sort_order", models.PositiveSmallIntegerField()),
                ("validator_regex", models.CharField(max_length=128, null=True)),
                ("vocabulary", models.JSONField(null=True)),
                ("label", models.CharField(max_length=128)),
                ("description", models.CharField(max_length=256)),
            ],
            options={
                "ordering": ["group__sort_order", "sort_order"],
            },
        ),
        migrations.CreateModel(
            name="AnnotationGroup",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128)),
                ("label", models.CharField(max_length=128)),
                ("sort_order", models.PositiveSmallIntegerField()),
            ],
            options={
                "ordering": ["sort_order"],
            },
        ),
        migrations.CreateModel(
            name="AnnotationValue",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("_value", models.JSONField(default=dict)),
                (
                    "entity",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="annotations",
                        to="flow.entity",
                    ),
                ),
                (
                    "field",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="values",
                        to="flow.annotationfield",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="AnnotationPreset",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128)),
                (
                    "fields",
                    models.ManyToManyField(
                        related_name="presets", to="flow.AnnotationField"
                    ),
                ),
                (
                    "permission_group",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="annotationpreset",
                        to="permissions.permissiongroup",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="annotationfield",
            name="group",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="fields",
                to="flow.annotationgroup",
            ),
        ),
        migrations.AddField(
            model_name="collection",
            name="annotation_fields",
            field=models.ManyToManyField(
                related_name="collection", to="flow.AnnotationField"
            ),
        ),
        migrations.AddConstraint(
            model_name="annotationfield",
            constraint=models.CheckConstraint(
                check=models.Q(("type__in", ["DATE", "DECIMAL", "INTEGER", "STRING"])),
                name="annotation_type",
            ),
        ),
        migrations.AddConstraint(
            model_name="annotationfield",
            constraint=models.UniqueConstraint(
                fields=("name", "group"), name="uniquetogether_name_group"
            ),
        ),
    ]



"""Workflow execution engines."""

from resolwe.flow.engine import BaseEngine


class BaseExecutionEngine(BaseEngine):
    """A workflow execution engine."""

    def evaluate(self, data):
        """Return the code needed to compute a given Data object."""
        raise NotImplementedError

    def get_expression_engine(self, name):
        """Return an expression engine by its name."""
        return self.manager.get_expression_engine(name)

    def get_output_schema(self, process):
        """Return any additional output schema for the process."""
        return []

    def discover_process(self, path):
        """Perform process discovery in given path.

        This method will be called during process registration and
        should return a list of dictionaries with discovered process
        schemas.
        """
        return []



"""Urls for Django Rest Framework."""

from django.urls import include, path
from rest_framework import routers

from .flow.views import (
    AnnotationFieldViewSet,
    AnnotationPresetViewSet,
    AnnotationValueViewSet,
    CollectionViewSet,
    DataViewSet,
    DescriptorSchemaViewSet,
    EntityViewSet,
    ProcessViewSet,
    RelationViewSet,
    StorageViewSet,
)
from .observers.views import BackgroundTaksViewSet
from .storage.views import UploadConfig, UploadCredentials

api_router = routers.DefaultRouter(trailing_slash=False)
api_router.register(r"collection", CollectionViewSet)
api_router.register(r"process", ProcessViewSet)
api_router.register(r"data", DataViewSet)
api_router.register(r"entity", EntityViewSet)
api_router.register(r"relation", RelationViewSet)
api_router.register(r"descriptorschema", DescriptorSchemaViewSet)
api_router.register(r"storage", StorageViewSet)
api_router.register(r"task", BackgroundTaksViewSet, basename="backgroundtask")
api_router.register(r"annotation_field", AnnotationFieldViewSet)
api_router.register(r"annotation_preset", AnnotationPresetViewSet)
api_router.register(r"annotation_value", AnnotationValueViewSet)


urlpatterns = [
    path("api/", include((api_router.urls, "resolwe-api"), namespace="resolwe-api")),
    path(
        "api/upload/credentials", UploadCredentials.as_view(), name="upload_credentials"
    ),
    path("api/upload/config", UploadConfig.as_view(), name="upload_config"),
    # TODO: The view below is deprecated and must be removed on 1. 1. 2025.
    path("api/upload_config", UploadConfig.as_view(), name="upload_config_deprecated"),
]
